#pragma once

#include <stack>

#include "SynTree/Mutator.h"
#include "SynTree/Visitor.h"

#include "SynTree/Initializer.h"
#include "SynTree/Statement.h"
#include "SynTree/Type.h"
#include "SynTree/Declaration.h"
#include "SynTree/Expression.h"
#include "SynTree/Constant.h"

#include "PassVisitor.proto.h"

//-------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Templated visitor type
// To use declare a PassVisitor< YOUR VISITOR TYPE >
// The visitor type should specify the previsit/postvisit for types that are desired.
//-------------------------------------------------------------------------------------------------------------------------------------------------------------------------
template< typename pass_type >
class PassVisitor final : public Visitor, public Mutator {
public:
	PassVisitor() = default;

	template< typename... Args >
	PassVisitor(Args &&... args)
		: pass( std::forward<Args>( args )... )
	{}

	virtual ~PassVisitor() = default;

	pass_type pass;

	virtual void visit( ObjectDecl *objectDecl ) override final;
	virtual void visit( FunctionDecl *functionDecl ) override final;
	virtual void visit( StructDecl *aggregateDecl ) override final;
	virtual void visit( UnionDecl *aggregateDecl ) override final;
	virtual void visit( EnumDecl *aggregateDecl ) override final;
	virtual void visit( TraitDecl *aggregateDecl ) override final;
	virtual void visit( TypeDecl *typeDecl ) override final;
	virtual void visit( TypedefDecl *typeDecl ) override final;
	virtual void visit( AsmDecl *asmDecl ) override final;

	virtual void visit( CompoundStmt *compoundStmt ) override final;
	virtual void visit( ExprStmt *exprStmt ) override final;
	virtual void visit( AsmStmt *asmStmt ) override final;
	virtual void visit( IfStmt *ifStmt ) override final;
	virtual void visit( WhileStmt *whileStmt ) override final;
	virtual void visit( ForStmt *forStmt ) override final;
	virtual void visit( SwitchStmt *switchStmt ) override final;
	virtual void visit( CaseStmt *caseStmt ) override final;
	virtual void visit( BranchStmt *branchStmt ) override final;
	virtual void visit( ReturnStmt *returnStmt ) override final;
	virtual void visit( TryStmt *tryStmt ) override final;
	virtual void visit( CatchStmt *catchStmt ) override final;
	virtual void visit( FinallyStmt *finallyStmt ) override final;
	virtual void visit( NullStmt *nullStmt ) override final;
	virtual void visit( DeclStmt *declStmt ) override final;
	virtual void visit( ImplicitCtorDtorStmt *impCtorDtorStmt ) override final;

	virtual void visit( ApplicationExpr *applicationExpr ) override final;
	virtual void visit( UntypedExpr *untypedExpr ) override final;
	virtual void visit( NameExpr *nameExpr ) override final;
	virtual void visit( CastExpr *castExpr ) override final;
	virtual void visit( AddressExpr *addressExpr ) override final;
	virtual void visit( LabelAddressExpr *labAddressExpr ) override final;
	virtual void visit( UntypedMemberExpr *memberExpr ) override final;
	virtual void visit( MemberExpr *memberExpr ) override final;
	virtual void visit( VariableExpr *variableExpr ) override final;
	virtual void visit( ConstantExpr *constantExpr ) override final;
	virtual void visit( SizeofExpr *sizeofExpr ) override final;
	virtual void visit( AlignofExpr *alignofExpr ) override final;
	virtual void visit( UntypedOffsetofExpr *offsetofExpr ) override final;
	virtual void visit( OffsetofExpr *offsetofExpr ) override final;
	virtual void visit( OffsetPackExpr *offsetPackExpr ) override final;
	virtual void visit( AttrExpr *attrExpr ) override final;
	virtual void visit( LogicalExpr *logicalExpr ) override final;
	virtual void visit( ConditionalExpr *conditionalExpr ) override final;
	virtual void visit( CommaExpr *commaExpr ) override final;
	virtual void visit( TypeExpr *typeExpr ) override final;
	virtual void visit( AsmExpr *asmExpr ) override final;
	virtual void visit( ImplicitCopyCtorExpr *impCpCtorExpr ) override final;
	virtual void visit( ConstructorExpr * ctorExpr ) override final;
	virtual void visit( CompoundLiteralExpr *compLitExpr ) override final;
	virtual void visit( UntypedValofExpr *valofExpr ) override final;
	virtual void visit( RangeExpr *rangeExpr ) override final;
	virtual void visit( UntypedTupleExpr *tupleExpr ) override final;
	virtual void visit( TupleExpr *tupleExpr ) override final;
	virtual void visit( TupleIndexExpr *tupleExpr ) override final;
	virtual void visit( MemberTupleExpr *tupleExpr ) override final;
	virtual void visit( TupleAssignExpr *assignExpr ) override final;
	virtual void visit( StmtExpr * stmtExpr ) override final;
	virtual void visit( UniqueExpr * uniqueExpr ) override final;

	virtual void visit( VoidType *basicType ) override final;
	virtual void visit( BasicType *basicType ) override final;
	virtual void visit( PointerType *pointerType ) override final;
	virtual void visit( ArrayType *arrayType ) override final;
	virtual void visit( FunctionType *functionType ) override final;
	virtual void visit( StructInstType *aggregateUseType ) override final;
	virtual void visit( UnionInstType *aggregateUseType ) override final;
	virtual void visit( EnumInstType *aggregateUseType ) override final;
	virtual void visit( TraitInstType *aggregateUseType ) override final;
	virtual void visit( TypeInstType *aggregateUseType ) override final;
	virtual void visit( TupleType *tupleType ) override final;
	virtual void visit( TypeofType *typeofType ) override final;
	virtual void visit( AttrType *attrType ) override final;
	virtual void visit( VarArgsType *varArgsType ) override final;
	virtual void visit( ZeroType *zeroType ) override final;
	virtual void visit( OneType *oneType ) override final;

	virtual void visit( SingleInit *singleInit ) override final;
	virtual void visit( ListInit *listInit ) override final;
	virtual void visit( ConstructorInit *ctorInit ) override final;

	virtual void visit( Subrange *subrange ) override final;

	virtual void visit( Constant *constant ) override final;

	virtual DeclarationWithType* mutate( ObjectDecl *objectDecl ) override final;
	virtual DeclarationWithType* mutate( FunctionDecl *functionDecl ) override final;
	virtual Declaration* mutate( StructDecl *aggregateDecl ) override final;
	virtual Declaration* mutate( UnionDecl *aggregateDecl ) override final;
	virtual Declaration* mutate( EnumDecl *aggregateDecl ) override final;
	virtual Declaration* mutate( TraitDecl *aggregateDecl ) override final;
	virtual TypeDecl* mutate( TypeDecl *typeDecl ) override final;
	virtual Declaration* mutate( TypedefDecl *typeDecl ) override final;
	virtual AsmDecl* mutate( AsmDecl *asmDecl ) override final;

	virtual CompoundStmt* mutate( CompoundStmt *compoundStmt ) override final;
	virtual Statement* mutate( ExprStmt *exprStmt ) override final;
	virtual Statement* mutate( AsmStmt *asmStmt ) override final;
	virtual Statement* mutate( IfStmt *ifStmt ) override final;
	virtual Statement* mutate( WhileStmt *whileStmt ) override final;
	virtual Statement* mutate( ForStmt *forStmt ) override final;
	virtual Statement* mutate( SwitchStmt *switchStmt ) override final;
	virtual Statement* mutate( CaseStmt *caseStmt ) override final;
	virtual Statement* mutate( BranchStmt *branchStmt ) override final;
	virtual Statement* mutate( ReturnStmt *returnStmt ) override final;
	virtual Statement* mutate( TryStmt *returnStmt ) override final;
	virtual Statement* mutate( CatchStmt *catchStmt ) override final;
	virtual Statement* mutate( FinallyStmt *catchStmt ) override final;
	virtual NullStmt* mutate( NullStmt *nullStmt ) override final;
	virtual Statement* mutate( DeclStmt *declStmt ) override final;
	virtual Statement* mutate( ImplicitCtorDtorStmt *impCtorDtorStmt ) override final;

	virtual Expression* mutate( ApplicationExpr *applicationExpr ) override final;
	virtual Expression* mutate( UntypedExpr *untypedExpr ) override final;
	virtual Expression* mutate( NameExpr *nameExpr ) override final;
	virtual Expression* mutate( AddressExpr *castExpr ) override final;
	virtual Expression* mutate( LabelAddressExpr *labAddressExpr ) override final;
	virtual Expression* mutate( CastExpr *castExpr ) override final;
	virtual Expression* mutate( UntypedMemberExpr *memberExpr ) override final;
	virtual Expression* mutate( MemberExpr *memberExpr ) override final;
	virtual Expression* mutate( VariableExpr *variableExpr ) override final;
	virtual Expression* mutate( ConstantExpr *constantExpr ) override final;
	virtual Expression* mutate( SizeofExpr *sizeofExpr ) override final;
	virtual Expression* mutate( AlignofExpr *alignofExpr ) override final;
	virtual Expression* mutate( UntypedOffsetofExpr *offsetofExpr ) override final;
	virtual Expression* mutate( OffsetofExpr *offsetofExpr ) override final;
	virtual Expression* mutate( OffsetPackExpr *offsetPackExpr ) override final;
	virtual Expression* mutate( AttrExpr *attrExpr ) override final;
	virtual Expression* mutate( LogicalExpr *logicalExpr ) override final;
	virtual Expression* mutate( ConditionalExpr *conditionalExpr ) override final;
	virtual Expression* mutate( CommaExpr *commaExpr ) override final;
	virtual Expression* mutate( TypeExpr *typeExpr ) override final;
	virtual Expression* mutate( AsmExpr *asmExpr ) override final;
	virtual Expression* mutate( ImplicitCopyCtorExpr *impCpCtorExpr ) override final;
	virtual Expression* mutate( ConstructorExpr *ctorExpr ) override final;
	virtual Expression* mutate( CompoundLiteralExpr *compLitExpr ) override final;
	virtual Expression* mutate( UntypedValofExpr *valofExpr ) override final;
	virtual Expression* mutate( RangeExpr *rangeExpr ) override final;
	virtual Expression* mutate( UntypedTupleExpr *tupleExpr ) override final;
	virtual Expression* mutate( TupleExpr *tupleExpr ) override final;
	virtual Expression* mutate( TupleIndexExpr *tupleExpr ) override final;
	virtual Expression* mutate( MemberTupleExpr *tupleExpr ) override final;
	virtual Expression* mutate( TupleAssignExpr *assignExpr ) override final;
	virtual Expression* mutate( StmtExpr * stmtExpr ) override final;
	virtual Expression* mutate( UniqueExpr * uniqueExpr ) override final;

	virtual Type* mutate( VoidType *basicType ) override final;
	virtual Type* mutate( BasicType *basicType ) override final;
	virtual Type* mutate( PointerType *pointerType ) override final;
	virtual Type* mutate( ArrayType *arrayType ) override final;
	virtual Type* mutate( FunctionType *functionType ) override final;
	virtual Type* mutate( StructInstType *aggregateUseType ) override final;
	virtual Type* mutate( UnionInstType *aggregateUseType ) override final;
	virtual Type* mutate( EnumInstType *aggregateUseType ) override final;
	virtual Type* mutate( TraitInstType *aggregateUseType ) override final;
	virtual Type* mutate( TypeInstType *aggregateUseType ) override final;
	virtual Type* mutate( TupleType *tupleType ) override final;
	virtual Type* mutate( TypeofType *typeofType ) override final;
	virtual Type* mutate( AttrType *attrType ) override final;
	virtual Type* mutate( VarArgsType *varArgsType ) override final;
	virtual Type* mutate( ZeroType *zeroType ) override final;
	virtual Type* mutate( OneType *oneType ) override final;

	virtual Initializer* mutate( SingleInit *singleInit ) override final;
	virtual Initializer* mutate( ListInit *listInit ) override final;
	virtual Initializer* mutate( ConstructorInit *ctorInit ) override final;

	virtual Subrange *mutate( Subrange *subrange ) override final;

	virtual Constant *mutate( Constant *constant ) override final;

private:
	template<typename node_type> void call_previsit ( node_type * node ) { previsit_impl ( pass, node, 0 ); }
	template<typename node_type> void call_postvisit( node_type * node ) { postvisit_impl( pass, node, 0 ); }

	template<typename node_type> void call_premutate ( node_type * node ) { premutate_impl( pass, node, 0 ); }
	template<typename return_type, typename node_type> return_type call_postmutate ( node_type * node ) { return postmutate_impl<return_type>( pass, node, 0 ); }

	void call_beginScope() { begin_scope_impl( pass, 0 ); }
	void call_endScope  () { end_scope_impl  ( pass, 0 ); }

	void set_env( TypeSubstitution * env ) { set_env_impl( pass, env, 0); }

	void visitStatementList( std::list< Statement* > &statements );
	void mutateStatementList( std::list< Statement* > &statements );

	Statement * visitStatement( Statement * stmt );
	Statement * mutateStatement( Statement * stmt );

	void visitExpression( Expression * expr );
	Expression * mutateExpression( Expression * expr );


	TypeSubstitution ** 		get_env_ptr    () { return env_impl             ( pass, 0); }
	std::list< Statement* > * 	get_beforeStmts() { return stmtsToAddBefore_impl( pass, 0); }
	std::list< Statement* > * 	get_afterStmts () { return stmtsToAddAfter_impl ( pass, 0); }
};

#include "PassVisitor.impl.h"
